"""Types module."""

from collections import deque
from dataclasses import dataclass
from typing import Any, Generator, NewType

from jsonschema import ValidationError, Validator
from lsprotocol import types as lsp
from yaml import CollectionNode, Mark, Token

# We can probably do better, but that will do for now
YamlDocument = NewType("YamlDocument", dict[str, Any])
Schema = NewType("Schema", dict[str, Any])


@dataclass
class ParsedResult:
    """Token parsed result container."""

    tokens: list[Token]
    instance: YamlDocument
    nodes: CollectionNode


@dataclass
class CompleteParsedResult(ParsedResult):
    """Indicate a complete document parsing."""

    pass


@dataclass
class IncompleteParsedResult(ParsedResult):
    """Indicate an incomplete parsing of the document."""

    pass


@dataclass
class DocumentNode:
    """Document node."""

    value: str
    start: Mark
    end: Mark
    selection_end: Mark

    def contains(self, position: lsp.Position) -> bool:
        """Is position contained in node range?"""
        range_start_before = self.start.line < position.line or (
            self.start.line == position.line and self.start.column <= position.character
        )

        range_end_after = self.selection_end.line > position.line or (
            self.selection_end.line == position.line
            and self.selection_end.column >= position.character
        )

        return range_start_before and range_end_after


@dataclass
class IndexEntry:
    """Document index entry."""

    validator: Validator
    tokens: list[Token]
    instance: YamlDocument
    segments: dict[tuple[str, ...], DocumentNode]
    version: int | None


class MissingTypeCharmcraftValidator:
    """No op implementation.

    Used if charmcraft.yaml is missing the 'type' key or is set to 'bundle'.
    """

    def iter_errors(
        self, instance: Any, _schema: Any = None
    ) -> Generator[ValidationError, None, None]:
        """Lazily yield each of the validation errors in the given instance."""
        yield ValidationError(
            validator="required",
            path=deque([]),
            message="'type' key is mandatory and must be 'charm'",
            schema={},
        )


class MissingTypeSnapcraftValidator:
    """No op implementation.

    Used if snapcraft.yaml is missing the 'base' or 'build-base' key.
    """

    def iter_errors(
        self, instance: Any, _schema: Any = None
    ) -> Generator[ValidationError, None, None]:
        """Lazily yield each of the validation errors in the given instance."""
        yield ValidationError(
            validator="required",
            path=deque([]),
            message="Filling 'base' and/or 'build-base' key(s) is mandatory.",
            schema={},
        )



import sys
from textwrap import dedent

import pytest
import pytest_lsp
from lsprotocol import types
from pytest_lsp import ClientServerConfig, LanguageClient


@pytest_lsp.fixture(
    config=ClientServerConfig(
        server_command=[sys.executable, "src/craft_ls/server.py"]
    ),
)
async def client(lsp_client: LanguageClient):
    # Setup
    params = types.InitializeParams(capabilities=types.ClientCapabilities())
    await lsp_client.initialize_session(params)

    yield

    # Teardown
    await lsp_client.shutdown_session()


@pytest.mark.asyncio
async def test_diagnostic_on_open(client: LanguageClient):
    """Ensure that the server implements diagnostics correctly."""
    # Given
    test_uri = "file:///path/to/snapcraft.yaml"
    text_content = dedent(
        """
        name: my_snap
        parts:
            foo:
                name: name
        """
    )

    # When
    client.text_document_did_open(
        params=types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=test_uri,
                language_id="yaml",
                version=1,
                text=text_content,
            )
        )
    )
    await client.wait_for_notification(types.TEXT_DOCUMENT_PUBLISH_DIAGNOSTICS)

    # Then
    assert (diagnostics := client.diagnostics.get(test_uri, []))
    assert any("is mandatory" in diagnostic.message for diagnostic in diagnostics)



"""Parser-validator core logic."""

import logging
import re
from collections import deque
from importlib.resources import files
from itertools import chain
from typing import Iterable, cast

import jsonref
import yaml
from jsonpath_ng import parse
from jsonschema import Draft202012Validator, ValidationError
from jsonschema.exceptions import relevance
from jsonschema.protocols import Validator
from jsonschema.validators import validator_for
from lsprotocol import types as lsp
from more_itertools import peekable
from referencing import Registry, Resource
from referencing.jsonschema import DRAFT202012
from yaml.events import (
    DocumentEndEvent,
    Event,
    MappingEndEvent,
    MappingStartEvent,
    SequenceEndEvent,
    SequenceStartEvent,
    StreamEndEvent,
)
from yaml.scanner import ScannerError
from yaml.tokens import (
    BlockEndToken,
    BlockMappingStartToken,
    BlockSequenceStartToken,
    KeyToken,
    ScalarToken,
    Token,
    ValueToken,
)

from craft_ls.types_ import (
    CompleteParsedResult,
    DocumentNode,
    IncompleteParsedResult,
    MissingTypeCharmcraftValidator,
    MissingTypeSnapcraftValidator,
    ParsedResult,
    Schema,
    YamlDocument,
)

SOURCE = "craft-ls"
FILE_TYPES = ["snapcraft", "rockcraft", "charmcraft"]
MISSING_DESC = "No description to display"
DEFAULT_RANGE = lsp.Range(
    start=lsp.Position(line=0, character=0),
    end=lsp.Position(line=0, character=0),
)
SPECIAL_SYMBOL_PARENTS = {"parts", "apps", "services"}

logger = logging.getLogger(__name__)

default_validators: dict[str, Validator] = {}
charmcraft_registry: Registry
for file_type in FILE_TYPES:
    schema_str = files("craft_ls.schemas").joinpath(f"{file_type}.json").read_text()
    schema = jsonref.loads(
        files("craft_ls.schemas").joinpath(f"{file_type}.json").read_text()
    )
    default_validators[file_type] = validator_for(schema)(schema)

    if file_type == "charmcraft":
        schema = Resource.from_contents(
            jsonref.loads(schema_str), default_specification=DRAFT202012
        )
        charmcraft_registry = schema @ Registry()

    if file_type == "snapcraft":
        schema = Resource.from_contents(
            jsonref.loads(schema_str), default_specification=DRAFT202012
        )
        snapcraft_registry = schema @ Registry()


def get_validator_and_parse(  # noqa: C901
    file_stem: str, instance_document: str
) -> tuple[Validator, ParsedResult] | None:
    """Get the most appropriate validator for the current document."""
    if file_stem not in FILE_TYPES:
        return None

    scanned_tokens = parse_tokens(instance_document)

    if file_stem == "rockcraft":
        return default_validators[file_stem], scanned_tokens

    elif file_stem == "snapcraft":
        base = scanned_tokens.instance.get("base", None)
        build_base = scanned_tokens.instance.get("build-base", None)
        validator: Draft202012Validator | MissingTypeSnapcraftValidator
        match base, build_base:
            case "core22", _:
                validator = Draft202012Validator(
                    schema=snapcraft_registry.resolver()
                    .lookup("urn:snapcraft:core22")
                    .contents
                )
            case "core24", _:
                validator = Draft202012Validator(
                    schema=snapcraft_registry.resolver()
                    .lookup("urn:snapcraft:core24")
                    .contents
                )
            case "bare", "core22":
                validator = Draft202012Validator(
                    schema=snapcraft_registry.resolver()
                    .lookup("urn:snapcraft:bare22")
                    .contents
                )
            case "bare", "core24":
                validator = Draft202012Validator(
                    schema=snapcraft_registry.resolver()
                    .lookup("urn:snapcraft:bare24")
                    .contents
                )
            case _, "core22":
                validator = Draft202012Validator(
                    schema=snapcraft_registry.resolver()
                    .lookup("urn:snapcraft:base22")
                    .contents
                )
            case _, "core24":
                validator = Draft202012Validator(
                    schema=snapcraft_registry.resolver()
                    .lookup("urn:snapcraft:base24")
                    .contents
                )
            case _, "devel":
                validator = Draft202012Validator(
                    schema=snapcraft_registry.resolver()
                    .lookup("urn:snapcraft:devel")
                    .contents
                )

            case _:
                validator = MissingTypeSnapcraftValidator()

        return cast(Validator, validator), scanned_tokens

    else:
        # by elimination, file_stem is charmcraft
        if scanned_tokens.instance.get("type") != "charm":
            return cast(Validator, MissingTypeCharmcraftValidator()), scanned_tokens

        validator = Draft202012Validator(
            schema=charmcraft_registry.resolver()
            .lookup("urn:charmcraft:platformcharm")
            .contents
        )
    return cast(Validator, validator), scanned_tokens


def parse_tokens(instance_document: str) -> ParsedResult:
    """Scan the document for yaml tokens."""
    tokens = []
    tokens_iter = yaml.scan(instance_document)

    try:
        for event in tokens_iter:
            tokens.append(event)
    except ScannerError:
        instance, events = robust_load(instance_document)
        nodes = yaml.compose(events)
        return IncompleteParsedResult(tokens=tokens, instance=instance, nodes=nodes)

    instance = cast(YamlDocument, yaml.safe_load(instance_document))
    nodes = yaml.compose(instance_document)
    return CompleteParsedResult(tokens=tokens, instance=instance, nodes=nodes)


def robust_load(instance_document: str) -> tuple[YamlDocument, list[Event]]:
    """Parse the valid portion of the stream and construct a Python object."""
    events = []
    events_iter = yaml.parse(instance_document)

    closing_sequence: deque[Event] = deque()

    try:
        for event in events_iter:
            match event:
                case MappingStartEvent():
                    closing_sequence.append(MappingEndEvent())
                case SequenceStartEvent():
                    closing_sequence.append(SequenceEndEvent())
                case MappingEndEvent():
                    closing_sequence.pop()
                case SequenceEndEvent():
                    closing_sequence.pop()

            events.append(event)
    except ScannerError:
        closing_sequence.extendleft([DocumentEndEvent(), StreamEndEvent()])

    truncated_file = yaml.emit(events + list(reversed(closing_sequence)))
    return cast(YamlDocument, yaml.safe_load(truncated_file)), truncated_file


def segmentize_nodes(
    root: yaml.CollectionNode,
) -> list[tuple[tuple[str, ...], DocumentNode]]:
    """Flatten graph into path segments."""
    segments: list[tuple[tuple[str, ...], DocumentNode]] = []
    nodes = list(root.value)

    for node_pair in nodes:
        segments.extend(_do_segmentize_nodes(*node_pair))

    return segments


def _do_segmentize_nodes(
    first: yaml.CollectionNode,
    second: yaml.CollectionNode,
    prefix: tuple[str, ...] | None = None,
) -> list[tuple[tuple[str, ...], DocumentNode]]:
    """Recursive node segmentation.

    Craft tools don't usually go over three levels, so we don't have to worry about recursion limits.
    """
    segments = []
    prefix = prefix or ()

    match second:
        case yaml.ScalarNode(end_mark=selection_end):
            current_node = DocumentNode(
                value=first.value,
                start=first.start_mark,
                end=first.end_mark,
                selection_end=selection_end,
            )
            segments.append((prefix + (str(first.value),), current_node))

        case yaml.MappingNode(end_mark=selection_end, value=children):
            current_node = DocumentNode(
                value=first.value,
                start=first.start_mark,
                end=first.end_mark,
                selection_end=selection_end,
            )
            segments.append((prefix + (str(first.value),), current_node))
            segments.extend(
                list(
                    chain.from_iterable(
                        [
                            _do_segmentize_nodes(
                                child[0], child[1], prefix=prefix + (str(first.value),)
                            )
                            for child in children
                        ]
                    )
                )
            )

        case yaml.SequenceNode(end_mark=selection_end):
            print(selection_end.line)
            current_node = DocumentNode(
                value=first.value,
                start=first.start_mark,
                end=first.end_mark,
                selection_end=selection_end,
            )
            segments.append((prefix + (str(first.value),), current_node))
        case other:
            logger.error(other)

    return segments


def get_diagnostics(
    validator: Validator,
    instance: YamlDocument,
    segments: dict[tuple[str, ...], DocumentNode],
) -> list[lsp.Diagnostic]:
    """Validate a document against its schema."""
    diagnostics = []

    for error in validator.iter_errors(instance):
        if error.context:
            error = sorted(error.context, key=relevance)[0]

        match error:
            case ValidationError(
                validator="additionalProperties", absolute_path=path, message=message
            ):
                ranges = [DEFAULT_RANGE]
                pattern = r"\((?P<keys>.*) (was|were) unexpected\)"
                if (match := re.search(pattern, message or "")) and (
                    keys := match.group("keys")
                ):
                    keys_cleaned = [key.strip(" '") for key in keys.split(",")]
                    ranges = [
                        get_diagnostic_range(segments, list(path) + [key])
                        for key in keys_cleaned
                    ]

                for range_ in ranges:
                    diagnostics.append(
                        lsp.Diagnostic(
                            message=message,
                            severity=lsp.DiagnosticSeverity.Error,
                            range=range_,
                            source=SOURCE,
                        )
                    )

            case ValidationError(
                validator="required",
                absolute_path=path,
                message=message,
            ):
                pattern = "'(?P<key>.*)' key is mandatory"
                if path:
                    range_ = get_diagnostic_range(segments, path)
                elif (match := re.search(pattern, message or "")) and (
                    key := match.group("key")
                ):
                    # Const errore can be wrongfully handled here with an empty path,
                    # so we try to handle those cases gracefully
                    range_ = get_diagnostic_range(segments, list(path) + [key])
                else:
                    range_ = DEFAULT_RANGE

                diagnostics.append(
                    lsp.Diagnostic(
                        message=message,
                        severity=lsp.DiagnosticSeverity.Error,
                        range=range_,
                        source=SOURCE,
                    )
                )

            case ValidationError(absolute_path=path, message=str(message)):
                path = cast(list[str], path)
                range_ = get_diagnostic_range(segments, path) if path else DEFAULT_RANGE

                diagnostics.append(
                    lsp.Diagnostic(
                        message=message,
                        severity=lsp.DiagnosticSeverity.Error,
                        range=range_,
                        source=SOURCE,
                    )
                )

            case error:
                # yet to implement
                logger.debug(error.message)

    return diagnostics


def get_diagnostic_range(
    document_segments: dict[tuple[str, ...], DocumentNode], diag_segments: Iterable[str]
) -> lsp.Range:
    """Link the validation error to the position in the original document."""
    if (
        not diag_segments
        or (error_node := document_segments.get(tuple(diag_segments))) is None
    ):
        return DEFAULT_RANGE

    range = lsp.Range(
        start=lsp.Position(
            line=error_node.start.line,
            character=error_node.start.column,
        ),
        end=lsp.Position(
            line=error_node.selection_end.line,
            character=error_node.selection_end.column,
        ),
    )
    return range


def sanatize_key(key: str) -> str:
    """Sanatize key."""
    return re.sub(r"""['"\\*]""", "", key)


def get_description_from_path(path: Iterable[str | int], schema: Schema) -> str:
    """Given an element path, get its description."""
    # The first part of the query must always be a perfect match according to all
    # schemas. It's also better for performance.
    head, *tail = path
    query = f"$.properties.{sanatize_key(str(head))}"
    if tail:
        sub_query = "..".join(
            [
                f"'{sanatize_key(str(p))}'|additionalProperties|patternProperties"
                for p in tail
            ]
        )
        query = f"{query}..{sub_query}"
    query = f"{query}.description|title"
    parser = parse(query)
    candidates = parser.find(schema)

    if candidates:
        return str(candidates[0].value).capitalize()
    else:
        return MISSING_DESC


def get_exact_cursor_path(position: lsp.Position, tokens: list[Token]) -> deque[str]:  # noqa: C901
    """Get the exact path to the cursor position."""
    current_path: deque[str] = deque()
    iterator = peekable(tokens)
    last_scalar_token: str = ""
    previous: Token | None = None
    token: Token | None = None
    next_token: Token | None = None

    for token in iterator:
        next_token = iterator.peek(None)
        early_stop = (
            not next_token
            or next_token.start_mark.line > position.line
            or (
                next_token.start_mark.line >= position.line
                and next_token.start_mark.column >= position.character
            )
        )
        if early_stop:
            break

        match token:
            case BlockMappingStartToken() | BlockSequenceStartToken():
                if last_scalar_token:
                    current_path.append(last_scalar_token)
            case BlockEndToken():
                if current_path:
                    current_path.pop()

            case KeyToken():
                last_scalar_token = ""

            case ScalarToken(value=value):
                if isinstance(previous, yaml.KeyToken):
                    last_scalar_token = value

        previous = token

    if (
        isinstance(token, (ValueToken, ScalarToken))
        and last_scalar_token
        and next_token
    ):
        current_path.append(last_scalar_token)

    return current_path


def get_node_path_from_token_position(
    position: lsp.Position, segments: dict[tuple[str, ...], DocumentNode]
) -> tuple[str, ...] | None:
    """Find the innermost node path corresponding to the current position."""
    for segment, node in reversed(segments.items()):
        if node.contains(position):
            return segment

    return None


def list_symbols(
    instance: YamlDocument, segments: dict[tuple[str, ...], DocumentNode]
) -> list[lsp.DocumentSymbol]:
    """List document symbols.

    We are only interested in keys up to the second level at most, so we don't need anything
    fancy here.
    """
    symbols = []
    for top_level_key in instance.keys():
        node = segments[(top_level_key,)]
        symbol = lsp.DocumentSymbol(
            name=node.value,
            kind=lsp.SymbolKind.Key,
            range=lsp.Range(
                start=lsp.Position(node.start.line, node.start.column),
                end=lsp.Position(node.end.line, node.end.column),
            ),
            selection_range=lsp.Range(
                start=lsp.Position(node.start.line, node.start.column),
                end=lsp.Position(node.selection_end.line, node.selection_end.column),
            ),
        )
        if top_level_key in SPECIAL_SYMBOL_PARENTS:
            children_symbols = []
            for second_level_key in instance[top_level_key].keys():
                child_node = segments[(top_level_key, second_level_key)]
                child_symbol = lsp.DocumentSymbol(
                    name=child_node.value,
                    kind=lsp.SymbolKind.Key,
                    range=lsp.Range(
                        start=lsp.Position(
                            child_node.start.line, child_node.start.column
                        ),
                        end=lsp.Position(child_node.end.line, child_node.end.column),
                    ),
                    selection_range=lsp.Range(
                        start=lsp.Position(
                            child_node.start.line, child_node.start.column
                        ),
                        end=lsp.Position(
                            child_node.selection_end.line,
                            child_node.selection_end.column,
                        ),
                    ),
                )
                children_symbols.append(child_symbol)

            symbol.children = children_symbols
        symbols.append(symbol)

    return symbols


def get_completion_items_from_path(
    segments: Iterable[str], schema: Schema, instance: YamlDocument
) -> list[lsp.CompletionItem]:
    """Get possible values for children nodes or enum values."""
    sub_instance = instance
    sub_schema = schema
    for segment in segments:
        sub_schema = sub_schema.get("properties", {}).get(segment, {})
        sub_instance = sub_instance.get(segment, {})

    already_present = (
        set(sub_instance.keys()) if isinstance(sub_instance, dict) else set()
    )
    items = []

    if "cons" in sub_schema.keys():
        items = [lsp.CompletionItem(label=str(key)) for key in [sub_schema["cons"]]]

    elif "enum" in sub_schema.keys():
        items = [
            lsp.CompletionItem(label=str(key)) for key in sub_schema["enum"] if key
        ]

    elif "properties" in sub_schema.keys():
        items = [
            lsp.CompletionItem(label=str(key))
            for key in set(sub_schema["properties"].keys()) - already_present
        ]
    return items



"""Use LSP capabilities directly using a command line interface.

Can be used to test the error parsing without the additional complexity of the server.
"""

import logging
import sys
from pathlib import Path

from lsprotocol import types as lsp

from craft_ls.core import get_diagnostics, get_validator_and_parse, segmentize_nodes
from craft_ls.types_ import ParsedResult

logging.basicConfig()


def check(file_name: str) -> None:
    """Report all violations for a file."""
    file = Path(file_name)

    diagnostics: list[lsp.Diagnostic] = []
    match get_validator_and_parse(file.stem, file.read_text()):
        case None:
            print(f"Cannot validate '{file}'", file=sys.stderr)
            pass

        case validator, ParsedResult(instance=instance, nodes=nodes):
            segments = segmentize_nodes(nodes)
            diagnostics.extend(get_diagnostics(validator, instance, dict(segments)))

    if diagnostics:
        for diag in diagnostics:
            print(f"{diag.range.start.line}: {diag.message}", file=sys.stderr)
        sys.exit(1)



import json
from collections import deque
from itertools import chain
from textwrap import dedent

import yaml
from hypothesis import assume, example, given
from hypothesis import strategies as st
from jsonschema.validators import validator_for
from lsprotocol import types as lsp

from craft_ls.core import (
    MISSING_DESC,
    get_description_from_path,
    get_diagnostic_range,
    get_diagnostics,
    get_exact_cursor_path,
    get_node_path_from_token_position,
    list_symbols,
    parse_tokens,
    segmentize_nodes,
)

# Adapted from json-schema.org
schema = json.loads(
    """
{
  "$schema": "https://json-schema.org/draft/2020-12/schema",
  "$id": "https://example.com/product.schema.json",
  "title": "Product",
  "description": "A product from Acme's catalog",
  "type": "object",
  "properties": {
    "productId": {
      "description": "The unique identifier for a product",
      "type": "integer"
    },
    "productName": {
      "description": "Name of the product",
      "type": "string"
    },
    "price": {
      "description": "The price of the product",
      "type": "object",
      "properties": {
        "amount": {
          "description": "The price amount",
          "type": "number",
          "exclusiveMinimum": 0
        },
        "currency": {
          "description": "The price currency",
          "type": "string"
        }
      }
    }
  },
  "required": ["productId", "productName", "price"],
  "additionalProperties": false
}
"""
)
validator = validator_for(schema)(schema)

document = """
# Some comment
productId: 001
productName: bar
price:
  amount: 50
  currency: euro
"""

parsed_document = parse_tokens(document)
document_segments = segmentize_nodes(parsed_document.nodes)


def test_get_description_first_level_ok() -> None:
    """Assert that we can get the description from the schema of a first-level key."""
    # Given
    item_path = ["price"]

    # When
    description = get_description_from_path(item_path, schema)

    # Then
    assert description == "The price of the product"


def test_get_description_nested_ok() -> None:
    """Assert that we can get the description from the schema of a non first-level key."""
    # Given
    item_path = ["price", "currency"]

    # When
    description = get_description_from_path(item_path, schema)

    # Then
    assert description == "The price currency"


@given(key=st.text(min_size=1))
@example("something_not_present")
def test_get_description_unknown_path(key: str) -> None:
    """Assert that we get a default message if the key is not in the schema."""
    # Given
    assume(key not in {"amount", "currency"})  # exclude the two counter examples
    item_path = ["price", key]

    # When
    description = get_description_from_path(item_path, schema)

    # Then
    assert description == MISSING_DESC


def test_get_node_path_from_position_first_level_ok() -> None:
    # Given
    # line is 2 because of initial newline after """ + comment in the document
    position = lsp.Position(2, 5)

    # When
    path = get_node_path_from_token_position(position, dict(document_segments))

    # Then
    assert path == ("productId",)


def test_get_node_path_from_position_nested_ok() -> None:
    # Given
    position = lsp.Position(5, 5)

    # When
    path = get_node_path_from_token_position(position, dict(document_segments))

    # Then
    assert path == ("price", "amount")


def test_get_node_path_from_outside_ko() -> None:
    # Given
    position = lsp.Position(1, 5)  # comment line

    # When
    path = get_node_path_from_token_position(position, dict(document_segments))

    # Then
    assert not path


def test_get_node_path_from_value_ok() -> None:
    # Given
    position = lsp.Position(4, 10)  # to the right of "price"

    # When
    path = get_node_path_from_token_position(position, dict(document_segments))

    # Then
    assert path == ("price",)


def test_get_cursor_path_from_token_ok() -> None:
    # Given
    position = lsp.Position(4, 3)  # inside "price", current path should be root

    # When
    path = get_exact_cursor_path(position, parsed_document.tokens)

    # Then
    assert path == deque([])


def test_get_cursor_path_from_nested_key_ok() -> None:
    # Given
    position = lsp.Position(5, 4)  # inside "amount", current path should be "price"

    # When
    path = get_exact_cursor_path(position, parsed_document.tokens)

    # Then
    assert path == deque(["price"])


def test_get_cursor_path_from_value_ok() -> None:
    # Given
    position = lsp.Position(3, 15)  # inside "bar", current path should be "productName"

    # When
    path = get_exact_cursor_path(position, parsed_document.tokens)

    # Then
    assert path == deque(["productName"])


def test_values_are_not_flagged() -> None:
    """Check that we flag the correct token: a key, not a value."""
    # Given
    # With the following document, an error on "productName" should
    # be matched with the third line, not the second one
    document = dedent(
        """
        # Some comment
        productId: productName
        productName: InvalidValue
        price:
          amount: 50
          currency: euro
        """
    )
    segments = dict(segmentize_nodes(parse_tokens(document).nodes))
    parsed_document.nodes

    # When
    range_ = get_diagnostic_range(segments, ["productName"])

    # Then
    assert range_.start.line == 3


def test_multiple_unexpected_keys() -> None:
    # Given
    # With the following document, we expect two diagnostics
    document = dedent(
        """
        # Some comment
        productId: 001
        productName: name
        price:
          amount: 50
          currency: euro
        foo: bar
        baz: buz
        """
    )
    segments = dict(segmentize_nodes(parse_tokens(document).nodes))

    # When
    diagnostics = get_diagnostics(validator, yaml.safe_load(document), segments)

    # Then
    assert len(diagnostics) == 2
    assert all("unexpected" in diag.message for diag in diagnostics)


def test_list_symbols_correct_levels() -> None:
    """We only expect 1st level keys and 2nd level keys for specific 1st keys.

    Using the document below, we expect 3 1st level keys and one child key.
    """
    # Given
    document = dedent(
        """
        # Some comment
        foo: bar
        parts:
            included-key:
                all: right
        unrelated:
            non-included-key: 50
        """
    )
    segments = dict(segmentize_nodes(yaml.compose(document)))

    # When
    symbols = list_symbols(yaml.safe_load(document), segments)

    # Then
    assert len(symbols) == 3
    children = list(
        chain.from_iterable([s.children for s in symbols if s.children is not None])
    )
    assert len(children) == 1
    assert children[0].name == "included-key"



"""Application module."""

import argparse
import logging

logger = logging.getLogger(__name__)


def main() -> None:
    """Application entrypoint."""
    from craft_ls import __version__

    parser = argparse.ArgumentParser(prog="craft-ls")
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
    )
    subparsers = parser.add_subparsers()
    parser_check = subparsers.add_parser(
        "check",
        help="Check a file and report violations",
    )
    parser_check.add_argument("file", help="File to check")
    parser_check.set_defaults(which="check")

    args = parser.parse_args()

    if not getattr(args, "which", ""):
        serve()
    elif args.which == "check":
        check(args.file)


def serve() -> None:
    """Run language server."""
    from craft_ls import server

    logger.info("Starting Craft-ls")
    server.start()


def check(file: str) -> None:
    """Run ad-hoc check."""
    from craft_ls.cli import check

    check(file)


if __name__ == "__main__":
    main()



"""Define the language server features."""

import logging
from pathlib import Path
from textwrap import shorten
from typing import cast

from lsprotocol import types as lsp
from pygls.lsp.server import LanguageServer

from craft_ls import __version__
from craft_ls.core import (
    get_completion_items_from_path,
    get_description_from_path,
    get_diagnostics,
    get_exact_cursor_path,
    get_node_path_from_token_position,
    get_validator_and_parse,
    list_symbols,
    segmentize_nodes,
)
from craft_ls.settings import IS_DEV_MODE
from craft_ls.types_ import IndexEntry, ParsedResult, Schema

MSG_SIZE = 79

logger = logging.getLogger(__name__)


class CraftLanguageServer(LanguageServer):
    """*craft tools language server."""

    def __init__(
        self,
        name: str,
        version: str,
        text_document_sync_kind: lsp.TextDocumentSyncKind = lsp.TextDocumentSyncKind.Incremental,
        notebook_document_sync: lsp.NotebookDocumentSyncOptions | None = None,
    ) -> None:
        super().__init__(
            name,
            version,
            text_document_sync_kind,
            notebook_document_sync,
        )
        self.index: dict[str, IndexEntry | None] = {}

    def parse_file(self, file_uri: str) -> IndexEntry | None:
        """Parse a document into tokens, nodes and whatnot.

        The result is cached so we can access it in endpoints.
        """
        document = self.workspace.get_text_document(file_uri)
        match get_validator_and_parse(Path(file_uri).stem, document.source):
            case None:
                self.index[file_uri] = None

            case validator, ParsedResult(tokens, instance, nodes):
                segments_nodes = segmentize_nodes(nodes)
                self.index[file_uri] = IndexEntry(
                    validator, tokens, instance, dict(segments_nodes), document.version
                )

        return self.index[file_uri]

    def get_or_update_index(self, file_uri: str) -> IndexEntry | None:
        """Re-parse document if needed."""
        current_version = self.workspace.get_text_document(file_uri).version
        entry = self.index.get(file_uri)
        match entry:
            case IndexEntry(version=cached_version) as cached:
                if not cached_version or cached_version != current_version:
                    return self.parse_file(
                        file_uri,
                    )
                return cached
            case None:
                return None


server = CraftLanguageServer(
    name="craft-ls",
    version=__version__,
)


def shorten_messages(diagnostics: list[lsp.Diagnostic]) -> None:
    """Shorten diagnostics messages to better fit an editor view."""
    for diagnostic in diagnostics:
        diagnostic.message = shorten(diagnostic.message, MSG_SIZE)


@server.feature(lsp.TEXT_DOCUMENT_DID_OPEN)
def on_opened(ls: CraftLanguageServer, params: lsp.DidOpenTextDocumentParams) -> None:
    """Parse each document when it is opened."""
    uri = params.text_document.uri
    diagnostics = (
        [
            lsp.Diagnostic(
                message=f"Running craft-ls {__version__}.",
                range=lsp.Range(
                    start=lsp.Position(line=0, character=0),
                    end=lsp.Position(line=0, character=0),
                ),
                severity=lsp.DiagnosticSeverity.Information,
            )
        ]
        if IS_DEV_MODE
        else []
    )

    match ls.parse_file(uri):
        case IndexEntry(
            validator, instance=instance, segments=segments, version=version
        ):
            diagnostics.extend(get_diagnostics(validator, instance, segments))

        case _:
            return

    shorten_messages(diagnostics)
    if diagnostics:
        server.text_document_publish_diagnostics(
            lsp.PublishDiagnosticsParams(
                uri=uri, version=version, diagnostics=diagnostics
            )
        )


@server.feature(lsp.TEXT_DOCUMENT_DID_CHANGE)
def on_changed(ls: CraftLanguageServer, params: lsp.DidOpenTextDocumentParams) -> None:
    """Parse each document when it is changed."""
    uri = params.text_document.uri
    diagnostics = []

    match ls.parse_file(uri):
        case IndexEntry(
            validator, instance=instance, segments=segments, version=version
        ):
            diagnostics.extend(get_diagnostics(validator, instance, segments))

        case _:
            return

    shorten_messages(diagnostics)
    server.text_document_publish_diagnostics(
        lsp.PublishDiagnosticsParams(uri=uri, version=version, diagnostics=diagnostics)
    )


@server.feature(lsp.TEXT_DOCUMENT_HOVER)
def hover(ls: CraftLanguageServer, params: lsp.HoverParams) -> lsp.Hover | None:
    """Get item description on hover."""
    pos = params.position
    uri = params.text_document.uri

    match ls.get_or_update_index(uri):
        case IndexEntry(validator_found, segments=segments):
            validator = validator_found

        case _:
            return None

    if not (path := get_node_path_from_token_position(position=pos, segments=segments)):
        return None

    description = get_description_from_path(
        path=path, schema=cast(Schema, validator.schema)
    )

    return lsp.Hover(
        contents=lsp.MarkupContent(
            kind=lsp.MarkupKind.Markdown,
            value=description,
        ),
        range=lsp.Range(
            start=lsp.Position(line=pos.line, character=0),
            end=lsp.Position(line=pos.line + 1, character=0),
        ),
    )


@server.feature(lsp.TEXT_DOCUMENT_DOCUMENT_SYMBOL)
def document_symbol(
    ls: CraftLanguageServer, params: lsp.DocumentSymbolParams
) -> list[lsp.DocumentSymbol]:
    """Return all the symbols defined in the given document."""
    uri = params.text_document.uri
    symbols_results: list[lsp.DocumentSymbol] = []

    match ls.get_or_update_index(uri):
        case IndexEntry(instance=instance, segments=segments):
            symbols_results = list_symbols(instance, segments)

    return symbols_results


@server.feature(
    lsp.TEXT_DOCUMENT_COMPLETION, lsp.CompletionOptions(trigger_characters=[" "])
)
def completions(
    ls: CraftLanguageServer, params: lsp.CompletionParams
) -> lsp.CompletionList | None:
    """Suggest next element based on the document structure."""
    pos = params.position
    uri = params.text_document.uri
    items = []

    match ls.get_or_update_index(uri):
        case IndexEntry(validator_found, instance=instance, tokens=tokens):
            validator = validator_found

        case _:
            return None

    path = get_exact_cursor_path(pos, tokens)
    items = get_completion_items_from_path(
        segments=path, schema=cast(Schema, validator.schema), instance=instance
    )

    return lsp.CompletionList(is_incomplete=False, items=items)


def start() -> None:
    """Start the server."""
    server.start_io()


if __name__ == "__main__":
    start()



"""Tasks definition for the nox runner."""

from pathlib import Path
from urllib.request import urlopen

import nox

nox.options.default_venv_backend = "uv"
nox.options.reuse_venv = "yes"
nox.options.sessions = ["fmt", "lint"]

# TODO: Pin versions to tagged release for all tools
CHARM_VERSION = "4.0.1"
SNAP_VERSION = "551a793ae20241db56b7edf9049070a72f5f593b"
ROCK_VERSION = "6553ac9de239be758f5cf648eb65c580709c4fc5"
SNAPCRAFT_URL = f"https://raw.githubusercontent.com/canonical/snapcraft/{SNAP_VERSION}/schema/snapcraft.json"
ROCKCRAFT_URL = f"https://raw.githubusercontent.com/canonical/rockcraft/{ROCK_VERSION}/schema/rockcraft.json"
CHARMCRAFT_URL = f"https://raw.githubusercontent.com/canonical/charmcraft/refs/tags/{CHARM_VERSION}/schema/charmcraft.json"


@nox.session()
def fmt(session: nox.Session) -> None:
    """Format source code."""
    session.run(
        "uv",
        "sync",
        "--frozen",
        "--only-group",
        "fmt",
        env={"UV_PROJECT_ENVIRONMENT": session.virtualenv.location},
    )
    session.run("ruff", "format", "src", "tests")


@nox.session()
def lint(session: nox.Session) -> None:
    """Lint source code."""
    session.run(
        "uv",
        "sync",
        "--frozen",
        "--group",
        "lint",
        env={"UV_PROJECT_ENVIRONMENT": session.virtualenv.location},
    )
    session.run("ruff", "check", "--fix", "src")
    session.run("ruff", "check", "--fix", "tests")
    session.run("mypy", "src")


@nox.session()
def tests(session: nox.Session) -> None:
    """Run the unit tests."""
    session.run_install(
        "uv",
        "sync",
        "--frozen",
        "--group",
        "unit",
        env={"UV_PROJECT_ENVIRONMENT": session.virtualenv.location},
    )
    session.run("python", "-m", "pytest", *session.posargs)


@nox.session()
def build_schemas(session=nox.Session) -> None:
    """Fetch the upstream schemas and apply patches."""
    tmp_build = Path(".build")
    resources_build = Path("dev")
    src = Path("src")
    tmp_build.mkdir(exist_ok=True)

    # Snapcraft
    session.log("Getting latest snapcraft schema")

    with (
        (snap := tmp_build / "snapcraft_upstream.json").open(
            "w", encoding="utf-8"
        ) as f,
        urlopen(SNAPCRAFT_URL) as g,
    ):
        f.write(g.read().decode("utf-8"))

    for patch_file in sorted(resources_build.glob("snap_*"), reverse=True):
        session.run("patch", snap, "-i", patch_file, external=True)

    # Rockcraft
    session.log("Getting latest rockcraft schema")

    with (
        (rock := tmp_build / "rockcraft_upstream.json").open(
            "w", encoding="utf-8"
        ) as f,
        urlopen(ROCKCRAFT_URL) as g,
    ):
        f.write(g.read().decode("utf-8"))

    for patch_file in sorted(resources_build.glob("rock_*"), reverse=True):
        session.run("patch", rock, "-i", patch_file, external=True)

    # Charmcraft
    session.log("Getting latest charmcraft schema")

    with (
        (charm := tmp_build / "charmcraft_upstream.json").open(
            "w", encoding="utf-8"
        ) as f,
        urlopen(CHARMCRAFT_URL) as g,
    ):
        f.write(g.read().decode("utf-8"))

    for patch_file in sorted(resources_build.glob("charm_*"), reverse=True):
        session.run("patch", charm, "-i", patch_file, external=True)

    session.log("Replacing schemas")
    snap.replace(src / "craft_ls/schemas/snapcraft.json")
    rock.replace(src / "craft_ls/schemas/rockcraft.json")
    charm.replace(src / "craft_ls/schemas/charmcraft.json")
